// Dictionary.hpp - part of Word Replacing Transformation by P.Skibinski

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define DICTNAME_EXT ".dic"
#define DICTNAME "wrt-"
#define SHORT_DICTNAME "wrt-short-"

#define HASH_TABLE_SIZE		(1<<21)
#define HASH_DOUBLE_MULT	37
#define HASH_MULT			23
int word_hash[HASH_TABLE_SIZE];

int sizeDict=0;
unsigned char** dict=NULL;
unsigned char* dictlen=NULL;
unsigned char* dictmem;

int ngram_hash[256][256];




#define CHARSET_COUNT		6

int lowerSet[CHARSET_COUNT][256];
int upperSet[CHARSET_COUNT][256];
int lowerSetRev[CHARSET_COUNT][256];
int upperSetRev[CHARSET_COUNT][256];
int freeUpper[CHARSET_COUNT],freeLower[CHARSET_COUNT];
int usedSet=0;

int reservedSet[256]; 
int addSymbols[256]; 
int sym2codeword[256]; 
int codeword2sym[256]; 
int value[256];

int dictionary,dict1size,dict2size,dict3size,dict4size,dict1plus2plus3,dict1plus2;
int bound4,bound3,dict123size,dict12size;



// convert upper string to lower
inline void toLower(unsigned char* s,int s_size)
{
	for (int i=0; i<s_size; i++)
		s[i]=TOLOWER(s[i]);
}


// convert lower string to upper
inline void toUpper(unsigned char* s,int s_size)
{
	for (int i=0; i<s_size; i++)
		s[i]=TOUPPER(s[i]); 
}

#define ORIGINAL_CHARSET(c)\
{\
	if (usedSet>0)\
	{\
		if (lowerSet[0][c]>0)\
			c=lowerSetRev[usedSet][lowerSet[0][c]];\
		else\
		if (upperSet[0][c]>0)\
			c=upperSetRev[usedSet][upperSet[0][c]];\
	}\
}

// make hash from string
inline unsigned int stringHash(const unsigned char *ptr, int len)
{
	unsigned int hash;
	for (hash = 0; len>0; len--, ptr++)
		hash = HASH_MULT * hash + *ptr;
 
	return hash&(HASH_TABLE_SIZE-1);
}


// check if word "s" does exist in the dictionary using hash "h" 
inline int checkHashExactly(const unsigned char* s,int s_size,int h)
{
	int i;

	i=word_hash[h];
	if (i>0)
	{
		if (dictlen[i]!=s_size || memcmp(dict[i],s,s_size)!=0)
		{
			i=word_hash[(h+s_size*HASH_DOUBLE_MULT)&(HASH_TABLE_SIZE-1)];
			if (i>0)
			{
				if (dictlen[i]!=s_size || memcmp(dict[i],s,s_size)!=0)
				{
					i=word_hash[(h+s_size*HASH_DOUBLE_MULT*HASH_DOUBLE_MULT)&(HASH_TABLE_SIZE-1)];
					if (i>0)
					{
						if (dictlen[i]!=s_size || memcmp(dict[i],s,s_size)!=0)
							i=-1;
					}
					else
						i=-1;
				}
			}
			else
				i=-1;
		}
	}
	else
		i=-1;

	if (i>=dictionary)
		i=-1;

	return i;
}

// check if word "s" (prefix of original word) does exist in the dictionary using hash "h" 
inline int checkHash(const unsigned char* s,int s_size,int h)
{
	int i;

	i=word_hash[h];
	if (i>0)
	{
		if (dictlen[i]>s_size || memcmp(dict[i],s,dictlen[i])!=0)
		{
			i=word_hash[(h+s_size*HASH_DOUBLE_MULT)&(HASH_TABLE_SIZE-1)];
			if (i>0)
			{
				if (dictlen[i]>s_size || memcmp(dict[i],s,dictlen[i])!=0)
				{
					i=word_hash[(h+s_size*HASH_DOUBLE_MULT*HASH_DOUBLE_MULT)&(HASH_TABLE_SIZE-1)];
					if (i>0)
					{
						if (dictlen[i]>s_size || memcmp(dict[i],s,dictlen[i])!=0)
							i=-1;
					}
					else
						i=-1;
				}
			}
			else
				i=-1;
		}
	}
	else
		i=-1;

	if (i>=dictionary)
		i=-1;

	return i;
}


// check if word "s" or prefix of word "s" does exist in the dictionary using hash "h" 
inline int findShorterWord(const unsigned char* s,int s_size)
{
	int ret;
	int i;
	int best;
	unsigned int hash;

	hash = 0;
	for (i=0; i<WORD_MIN_SIZE+tryShorterBound; i++)
		hash = HASH_MULT * hash + s[i];
 
	best=-1;
	for (i=WORD_MIN_SIZE+tryShorterBound; i<s_size; i++)
	{
		ret=checkHash(s,i,hash&(HASH_TABLE_SIZE-1));	
		if (ret>=0)
			best=ret;
		hash = HASH_MULT*hash + s[i];
	}

	return best;
}

inline int findShorterWordRev(const unsigned char* s,int s_size)
{
	int ret;
	int i;

	for (i=s_size-1; i>=WORD_MIN_SIZE+tryShorterBound; i--)
	{
		ret=checkHash(s+s_size-i,i,stringHash(s+s_size-i,i));	
		if (ret>=0)
			return ret;
	}

	return -1;
}

inline void encodeCodeWord_LZ(int i,FILE* fileout);
inline void encodeCodeWord(int i,FILE* fileout);

// encode word (should be lower case) using n-gram array (when word doesn't exist in the dictionary)
inline void encodeAsText(unsigned char* s,int s_size,FILE* fileout)
{
	int i,ngram;

	if (spaceBefore!=NONE)
	{
		if (spaceBefore==SPACE)
			ENCODE_PUTC(' ',fileout);
		spaceBefore=NONE;
	}

	if (usedSet>0)
	{
		for (i=0; i<s_size; i++)
		{
			ORIGINAL_CHARSET(s[i]);
			ENCODE_PUTC(s[i],fileout);
		}
	}
	else
	{
		ngram=0;
		for (i=0; i<s_size; )
		{
			if (IF_OPTION(OPTION_USE_NGRAMS))
				ngram=ngram_hash[s[i]][s[i+1]];

			if (ngram>0 && ngram<dict1size && preprocType!=LZ77)
			{
				encodeCodeWord(ngram,fileout);
				i+=2;
			}
			else
			{
				ENCODE_PUTC(s[i],fileout);
				i++;
			}
		}
	}
}


inline void encodeCodeWord_LZ(int i,FILE* fileout)
{
	int first,second,third,fourth;
	int no=i;

	first=i-1;

	if (first>=bound4)
	{
		first-=bound4;

		fourth=first/dict123size;
		first=first%dict123size;
		third=first/dict12size;		
		first=first%dict12size;
		second=first/dict1size;		
		first=first%dict1size;

		ENCODE_PUTC(sym2codeword[dict1size+fourth],fileout);

		ENCODE_PUTC(sym2codeword[dict1size+third],fileout);

		ENCODE_PUTC(sym2codeword[dict1size+second],fileout);

		ENCODE_PUTC(sym2codeword[first],fileout);
	}
	else
	if (first>=bound3)
	{
		first-=bound3;

		third=first/dict12size;		
		first=first%dict12size;
		second=first/dict1size;		
		first=first%dict1size;

		ENCODE_PUTC(sym2codeword[dict1size+third],fileout);

		ENCODE_PUTC(sym2codeword[dict1size+second],fileout);

		ENCODE_PUTC(sym2codeword[first],fileout);
	}
	else
		if (first>=dict1size)
		{
			first-=dict1size;

			second=first/dict1size;		
			first=first%dict1size;

			ENCODE_PUTC(sym2codeword[dict1size+second],fileout);

			ENCODE_PUTC(sym2codeword[first],fileout);
		}
		else
		{
			ENCODE_PUTC(sym2codeword[first],fileout);
		}

}



inline void encodeCodeWord(int i,FILE* fileout)
{
	int first,second,third,fourth;
	int no=i;


	first=i-1;

	if (first>=bound4)
	{
		first-=bound4;

		fourth=first/dict123size;
		first=first%dict123size;
		third=first/dict12size;		
		first=first%dict12size;
		second=first/dict1size;		
		first=first%dict1size;

		ENCODE_PUTC(sym2codeword[dict1plus2plus3+fourth],fileout);

		ENCODE_PUTC(sym2codeword[dict1plus2+third],fileout);

		ENCODE_PUTC(sym2codeword[dict1size+second],fileout);

		ENCODE_PUTC(sym2codeword[first],fileout);
	}
	else
	if (first>=bound3)
	{
		first-=bound3;

		third=first/dict12size;		
		first=first%dict12size;
		second=first/dict1size;		
		first=first%dict1size;

		ENCODE_PUTC(sym2codeword[dict1plus2+third],fileout);

		ENCODE_PUTC(sym2codeword[dict1size+second],fileout);

		ENCODE_PUTC(sym2codeword[first],fileout);
	}
	else
		if (first>=dict1size)
		{
			first-=dict1size;

			second=first/dict1size;		
			first=first%dict1size;

			ENCODE_PUTC(sym2codeword[dict1size+second],fileout);

			ENCODE_PUTC(sym2codeword[first],fileout);
		}
		else
		{
			ENCODE_PUTC(sym2codeword[first],fileout);
		}

}


// encode word "s" using dictionary
inline void encodeWord(FILE* fileout,unsigned char* s,int s_size,EWordType wordType)
{
	int i;
	int size=0;
	int flagToEncode=-1;

	if (s_size<1)
	{
		if (spaceBefore!=NONE)
		{
			if (spaceBefore==SPACE)
				ENCODE_PUTC(' ',fileout);
			spaceBefore=NONE;
		}
		return;
	}

	s[s_size]=0;

	if (wordType!=LOWERWORD)
	{
		if (IF_OPTION(OPTION_CAPTIAL_CONVERSION))
		{
		if (wordType==FIRSTUPPER)
		{
			flagToEncode=CHAR_FIRSTUPPER;
			s[0]=TOLOWER(s[0]);
		}
		else // wordType==UPPERWORD
		{
			flagToEncode=CHAR_UPPERWORD;
			toLower(s,s_size);
		}
		}
		else
			wordType=LOWERWORD;
	}
	

	if (IF_OPTION(OPTION_USE_DICTIONARY))
	{
		i=checkHashExactly(s,s_size,stringHash(s,s_size));


		if (i<0 && IF_OPTION(OPTION_TRY_SHORTER_WORD))
		{
			// try to find shorter version of word in dictionary
			i=findShorterWord(s,s_size);
			if (i>=0)
				size=dictlen[i];
			else
				if (!IF_OPTION(OPTION_SPACELESS_WORDS))
				{
					i=findShorterWordRev(s,s_size);
					if (i>=0)
					{
						if (wordType!=LOWERWORD)
						{
							ENCODE_PUTC(flagToEncode,fileout);
							if (IF_OPTION(OPTION_SPACE_AFTER_CC_FLAG))
								ENCODE_PUTC(' ',fileout);
							wordType=LOWERWORD;
						}
						
						s[s_size-dictlen[i]]=0;
						encodeAsText(s,s_size-dictlen[i],fileout);
						s+=dictlen[i];
						s_size-=dictlen[i];
					}
				}
		}
	}
	else
		i=-1;


	if (i>=0)
	{
		if (wordType!=LOWERWORD)
		{
			ENCODE_PUTC(flagToEncode,fileout);
			if (IF_OPTION(OPTION_SPACE_AFTER_CC_FLAG))
				ENCODE_PUTC(' ',fileout);
		}

		if (spaceBefore==NONE)
		{
			if (IF_OPTION(OPTION_SPACELESS_WORDS))
				ENCODE_PUTC(CHAR_NOSPACE,fileout);
		}
		else
			spaceBefore=NONE;

		if (preprocType==LZ77)
			encodeCodeWord_LZ(i,fileout);
		else
			encodeCodeWord(i,fileout);

		if (size>0)
			encodeAsText(s+size,s_size-size,fileout);
	}
	else
	{
		if (wordType!=LOWERWORD)
		{
			if (spaceBefore!=NONE)
			{
				if (spaceBefore==SPACE)
					ENCODE_PUTC(' ',fileout);
				spaceBefore=NONE;
			}

			ENCODE_PUTC(flagToEncode,fileout);
			if (IF_OPTION(OPTION_SPACE_AFTER_CC_FLAG))
				ENCODE_PUTC(' ',fileout);
		}
 
		encodeAsText(s,s_size,fileout);
	}
}


// decode word using dictionary
#define DECODE_WORD(dictNo,i)\
{\
		switch (dictNo)\
		{\
			case 4:\
				i+=bound4;\
				break;\
			case 3:\
				i+=bound3;\
				break;\
			case 2:\
				i+=dict1size;\
				break;\
		}\
\
\
		if (i>=0 && i<sizeDict)\
		{\
			i++;\
			s_size=dictlen[i];\
			memcpy(s,dict[i],s_size+1);\
		}\
		else\
		{\
			printf("File is corrupted!\n");\
			exit(2);\
		}\
}


inline int decodeCodeWord_LZ(FILE* file, unsigned char* s,int& c)
{
	int i,i2,dictNo,s_size;


	if (codeword2sym[c]<dict1size)
	{
		i=codeword2sym[c];
		dictNo=1;
		DECODE_WORD(dictNo, i);
		return s_size;
	}
	else
		i=codeword2sym[c]-dict1size;


	DECODE_GETC(c,file);


	if (codeword2sym[c]<dict1size)
	{
		i=i*dict1size+codeword2sym[c];
		dictNo=2;
		DECODE_WORD(dictNo, i);
		return s_size;
	}
	else
		i2=codeword2sym[c]-dict1size;


	DECODE_GETC(c,file);


	if (codeword2sym[c]<dict1size)
	{
		i=i*dict12size+i2*dict1size+codeword2sym[c];
		dictNo=3;
		DECODE_WORD(dictNo, i);
		return s_size;
	}
	else
		i=i*dict123size+i2*dict12size+(codeword2sym[c]-dict1size)*dict1size;


	DECODE_GETC(c,file);


	if (codeword2sym[c]<dict1size)
		i=i+codeword2sym[c];
	else 
		printf("error\n");



	dictNo=4;
	DECODE_WORD(dictNo, i);

	return s_size;
}

inline int decodeCodeWord(FILE* file, unsigned char* s,int& c)
{
	int i,dictNo,s_size;



	if (codeword2sym[c]<dict1size)
	{
		i=codeword2sym[c];
		dictNo=1;
		DECODE_WORD(dictNo, i);
		return s_size;
	}
	else
	if (codeword2sym[c]<dict1plus2)
		i=dict1size*(codeword2sym[c]-dict1size);
	else
	if (codeword2sym[c]<dict1plus2plus3)
		i=dict12size*(codeword2sym[c]-dict1plus2);
	else
		i=dict123size*(codeword2sym[c]-dict1plus2plus3);



	DECODE_GETC(c,file);



	if (codeword2sym[c]<dict1size)
	{
		i+=codeword2sym[c];
		dictNo=2;
		DECODE_WORD(dictNo, i);
		return s_size;
	}
	else
	if (codeword2sym[c]<dict1plus2)
		i+=dict1size*(codeword2sym[c]-dict1size);
	else
		i+=dict12size*(codeword2sym[c]-dict1plus2);

	DECODE_GETC(c,file);



	if (codeword2sym[c]<dict1size)
	{
		i+=codeword2sym[c];
		dictNo=3;
		DECODE_WORD(dictNo, i);
		return s_size;
	}
	else
	if (codeword2sym[c]<dict1plus2)
		i+=dict1size*(codeword2sym[c]-dict1size);


	DECODE_GETC(c,file);


	if (codeword2sym[c]<dict1size)
		i+=codeword2sym[c];
	else 
		printf("error\n");




	dictNo=4;
	DECODE_WORD(dictNo, i);

	return s_size;
}

unsigned char* loadDictionary(FILE* file,unsigned char* mem,int word_count)
{
	unsigned char* word;
	int i,j,c,collision,bound;

	collision=0;
	bound=sizeDict+word_count;


	while (!feof(file))
	{
		word=mem;
		do
		{
			c=getc(file);
			if (usedSet==CHARSET_COUNT-1)
			{
				if (lowerSet[0][c]>0)
					c=lowerSetRev[CHARSET_COUNT-1][lowerSet[0][c]];
			}

			word[0]=c;
			word++;
		}
		while (c>32);

		if (c==EOF)
			break;
		if (c=='\r') 
			c=getc(file);
		
		word[-1]=0;
		i=word-mem-1;
		
		dictlen[sizeDict]=i;
		dict[sizeDict]=mem;



		j=stringHash(mem,i);
		mem+=(i/4+1)*4;
		

		if (word_hash[j]!=0)
		{
			if (dictlen[sizeDict]!=dictlen[word_hash[j]] || memcmp(dict[sizeDict],dict[word_hash[j]],dictlen[sizeDict])!=0)
			{
				c=(j+i*HASH_DOUBLE_MULT)&(HASH_TABLE_SIZE-1);
				if (word_hash[c]!=0)
				{
					if (dictlen[sizeDict]!=dictlen[word_hash[c]] || memcmp(dict[sizeDict],dict[word_hash[c]],dictlen[sizeDict])!=0)
					{
						c=(j+i*HASH_DOUBLE_MULT*HASH_DOUBLE_MULT)&(HASH_TABLE_SIZE-1);
						if (word_hash[c]!=0)
						{
							collision++;
						}
						else
						{
							if (dictlen[sizeDict]==2)
								ngram_hash[dict[sizeDict][0]][dict[sizeDict][1]]=sizeDict;
							word_hash[c]=sizeDict++;
						}
					}
				}
				else
				{
					if (dictlen[sizeDict]==2)
						ngram_hash[dict[sizeDict][0]][dict[sizeDict][1]]=sizeDict;
					word_hash[c]=sizeDict++;
				}
			}
		}
		else
		{
			if (dictlen[sizeDict]==2)
				ngram_hash[dict[sizeDict][0]][dict[sizeDict][1]]=sizeDict;
			word_hash[j]=sizeDict++;
		}

		
		if (sizeDict>dictionary || sizeDict>=bound)
		{
			sizeDict--;
			break;
		}
		
	}

	if (collision>0)
		printf("warning: hash collisions=%d\n",collision);

	return mem;
}

int loadCharset(FILE* file,int& freeChar,int* charset,int* charsetRev,bool *joinCharsets=NULL)
{
	int c,res,mult;
	
	res=0;
	c=getc(file); 
	mult=100;

	while (c>32)
	{
		{
			if (joinCharsets)
				joinCharsets[c]=true;

			charsetRev[freeChar]=c;
			charset[c]=freeChar++;

		}

		res+=mult*value[c];
		mult--;

		c=getc(file);
	}

	if (c==13)
		c=getc(file); // skip CR+LF or LF

	return res;
}


void initializeCodeWords()
{
	int c,charsUsed,i;

	for (c=0; c<256; c++)
	{
		addSymbols[c]=0;
		codeword2sym[c]=0;
		sym2codeword[c]=0;
		reservedSet[c]=0;
	}

	for (c=BINARY_FIRST; c<=BINARY_LAST; c++)
		addSymbols[c]=1;

//	addSymbols[151]=0; // pause symbol


	if (IF_OPTION(OPTION_ADD_SYMBOLS_MISC))
	{
		addSymbols[35]=1;
		addSymbols[38]=1;
		addSymbols[60]=1;
	//	addSymbols[61]=1;
		addSymbols[62]=1;
		addSymbols[64]=1;
		addSymbols[94]=1;
		addSymbols[96]=1;
		addSymbols[123]=1;
		addSymbols[124]=1;
		addSymbols[125]=1;
		addSymbols[126]=1;
		addSymbols[127]=1;
	}

	if (IF_OPTION(OPTION_ADD_SYMBOLS_0_4))
		for (c=0; c<=4; c++)
			addSymbols[c]=1;

	if (IF_OPTION(OPTION_ADD_SYMBOLS_14_31))
		for (c=14; c<=31; c++)
			addSymbols[c]=1;



	for (c=0; c<256; c++)
	{
		if (IF_OPTION(OPTION_USE_DICTIONARY) && addSymbols[c]
			&& (lowerSet[usedSet][c]!=0))
			addSymbols[c]=0;

		if (IF_OPTION(OPTION_USE_DICTIONARY) && addSymbols[c]
			&& (upperSet[usedSet][c]!=0) && !IF_OPTION(OPTION_CAPTIAL_CONVERSION))
			addSymbols[c]=0;

		if ( (IF_OPTION(OPTION_USE_DICTIONARY) && addSymbols[c])
			|| c==CHAR_ESCAPE || c==CHAR_LOWERWORD || c==CHAR_FIRSTUPPER || c==CHAR_UPPERWORD || c==CHAR_CR_LF
			|| (IF_OPTION(OPTION_SPACELESS_WORDS) && c==CHAR_NOSPACE) )
			reservedSet[c]=1;
		else
			reservedSet[c]=0;
	}

	if (IF_OPTION(OPTION_ADD_SYMBOLS_A_Z) && IF_OPTION(OPTION_CAPTIAL_CONVERSION))
		for (c=NGRAM_FIRST; c<=NGRAM_LAST; c++)
			addSymbols[c]=1;


	if (IF_OPTION(OPTION_USE_DICTIONARY))
	{
		charsUsed=0;
		for (c=0; c<256; c++)
		{
			if (addSymbols[c])
			{
				codeword2sym[c]=charsUsed;
				sym2codeword[charsUsed]=c;
				charsUsed++;
			}
		}
		
		c=dict1size;
		
		if (preprocType!=LZ77)
		{
			dict2size=dict3size=dict4size=charsUsed/4;
			dict1size=charsUsed-dict4size*3;
			
			for (i=0; i<charsUsed/4; i++)
			{
				if (i*i*i*(charsUsed-i)>c)
				{
					dict1size=charsUsed-i*3;
					dict2size=i;
					dict3size=i;
					dict4size=i;
					break;
				}
			}
		}
		else
		{
			for (i=0; i<charsUsed/3; i++)
			{
				dict4size=0;
				dict2size=dict3size=charsUsed/3;
				dict1size=charsUsed-dict3size*2;
				
				if (i*i*(charsUsed-i)>c)
				{
					dict1size=charsUsed-i;
					dict2size=i;
					dict3size=i;
					dict4size=0;
					break;
				}
			}
		}
		
		if (preprocType==PAQ)
		{
			dict2size=32;
			dict3size=16;
			dict4size=16;
			dict1size=charsUsed-dict2size-dict3size-dict4size;
		}

		dictionary=(dict1size*dict2size*dict3size*dict4size+dict1size*dict2size*dict3size+dict1size*dict2size+dict1size);
		bound4=dict1size*dict2size*dict3size+dict1size*dict2size+dict1size;
		bound3=dict1size*dict2size+dict1size;
		dict123size=dict1size*dict2size*dict3size;
		dict12size=dict1size*dict2size;
		dict1plus2=dict1size+dict2size;
		dict1plus2plus3=dict1size+dict2size+dict3size;
		
		dict=(unsigned char**)calloc(sizeof(unsigned char*)*(dictionary+1),1);
		dictlen=(unsigned char*)calloc(sizeof(unsigned char)*(dictionary+1),1);
	}
}


// read dictionary from files to arrays
bool initialize(unsigned char* dictName,unsigned char* shortDictName,bool encoding)
{
//	printf("dictName=%s shortDictName=%s\n",dictName,shortDictName);

	int i,j,c,set[CHARSET_COUNT],fileLen;
	FILE* file,*file2;
	unsigned char* mem;

	dict=NULL;
	dictlen=NULL;
	dictmem=NULL;

	memset(&word_hash[0],0,HASH_TABLE_SIZE*sizeof(word_hash[0]));
	memset(lowerSet,0,sizeof(lowerSet));
	memset(upperSet,0,sizeof(upperSet));
	memset(lowerSetRev,0,sizeof(lowerSetRev));
	memset(upperSetRev,0,sizeof(upperSetRev));
	memset(ngram_hash,0,sizeof(ngram_hash));

	if (dictName==NULL && shortDictName==NULL)
	{
		initializeCodeWords();
		sizeDict=0;
		return true;
	}

	if (dictName==NULL && shortDictName!=NULL)
	{
		dictName=shortDictName;
		shortDictName=NULL;
	}



	if (IF_OPTION(OPTION_USE_DICTIONARY))
	{
		printf("- loading dictionary %s\n",dictName); 

		file=fopen((const char*)dictName,"rb");
		if (file==NULL)
		{
			printf("Can't open dictionary %s\n",dictName);
			return false;
		}

		fileLen=flen(file);
		fscanf(file,"%d",&dict1size);


		do { c=getc(file); } while (c>=32); if (c==13) c=getc(file); // skip CR+LF or LF


		for (i=0; i<CHARSET_COUNT; i++)
		{
			freeUpper[i]=1;
			set[i]=loadCharset(file,freeUpper[i],upperSet[i],upperSetRev[i]);
			freeLower[i]=1;
			set[i]+=loadCharset(file,freeLower[i],lowerSet[i],lowerSetRev[i]);
		}

		if (encoding)
		{
			j=0;
			for (i=1; i<CHARSET_COUNT-1; i++)
			{
				if (set[i]>set[j])
					j=i;
			}

			usedSet=j;

			if (set[usedSet]==0 && freeLower[CHARSET_COUNT-1]>1)
				usedSet=CHARSET_COUNT-1;
		}

	
		if (freeUpper[0]!=freeLower[0])
		{
			printf("Bad the first charset in %s (the 2nd and the 3rd line length are different)\n",dictName);
			fclose(file);
			return false;
		}

		if (freeUpper[usedSet]!=freeLower[usedSet] || freeLower[usedSet]!=freeLower[0])
		{
			printf("Bad the second charset in %s (the %dth and the %dth line length are different)\n",dictName,usedSet*2-2,usedSet*2-1);
			fclose(file);
			return false;
		}

		dictmem=(unsigned char*)calloc(fileLen*2,1);
		mem=dictmem;
		sizeDict=1;

		if (shortDictName)
		{
			file2=fopen((const char*)shortDictName,"rb");
			if (file2==NULL)
			{
				printf("Can't open dictionary %s\n",shortDictName);
				return false;
			}
				
			do c=getc(file2); while (c>=32); if (c==13) c=getc(file2); 
			for (i=0; i<CHARSET_COUNT; i++)
			{
				loadCharset(file2,freeUpper[i],upperSet[i],upperSetRev[i]);
				loadCharset(file2,freeLower[i],lowerSet[i],lowerSetRev[i]);
			}

			if (freeUpper[0]!=freeLower[0])
			{
				printf("Bad the first charset in %s (the 2nd and the 3rd line length are different)\n",shortDictName);
				fclose(file2);
				fclose(file);
				return false;
			}

			if (freeUpper[usedSet]!=freeLower[usedSet] || freeLower[usedSet]!=freeLower[0])
			{
				printf("Bad the second charset in %s (the %dth and the %dth line length are different)\n",dictName,usedSet*2-2,usedSet*2-1);
				fclose(file2);
				fclose(file);
				return false;
			}

			initializeCodeWords();

			if (dict==NULL || dictlen==NULL)
				return false;

			mem=loadDictionary(file2,mem,dictionary);
			fclose(file2);

		}
		else
		{
			initializeCodeWords();

			if (dict==NULL || dictlen==NULL)
				return false;
		}

		mem=loadDictionary(file,mem,dictionary);

		if (encoding && usedSet==CHARSET_COUNT-1)
		{
			memset(lowerSet,0,sizeof(lowerSet));
			memset(upperSet,0,sizeof(upperSet));
			memset(lowerSetRev,0,sizeof(lowerSetRev));
			memset(upperSetRev,0,sizeof(upperSetRev));
		}

		fclose(file);

		printf("- loaded dictionary %d/%d words\n",sizeDict,dictionary);
	}
	else
	{
		initializeCodeWords();
		sizeDict=0;
	}

	return true;
}


void WRT_deinitialize()
{
	if (dict)
		free(dict);
	if (dictlen)
		free(dictlen);
	if (dictmem)
		free(dictmem);
}
